"""
Provides a human readable info about OpenGL data types
including their actual byte size
"""

from OpenGL import GL

SIZE_OF_FLOAT = 4
SIZE_OF_DOUBLE = 8
SIZE_OF_INT = 4
SIZE_OF_SHORT = 2


class TypeInfo:
    """
    Information about a data types in OpenGL.

    Example: "GL_FLOAT_VEC3" is a GL.GL_FLOAT_VEC3 of byte size 12
    """
    def __init__(self, name, value, size):
        """
        :param name: Name of the type (ex 'GL_FLOAT_VEC4')
        :param value: Enum value in OpenGL
        :param size:  Size in bytes
        """
        self.name = name
        self.value = value
        self.size = size

    def __str__(self):
        return "<TypeInfo name={}, value={} size={}>".format(self.name, self.value, self.size)

    def __repr__(self):
        return str(self)


# Information about each GL enum type
TYPE_INFO = {
    # Floats
    GL.GL_FLOAT: TypeInfo("GL_FLOAT", GL.GL_FLOAT, SIZE_OF_FLOAT),
    GL.GL_FLOAT_VEC2: TypeInfo("GL_FLOAT_VEC2", GL.GL_FLOAT_VEC2, SIZE_OF_FLOAT * 2),
    GL.GL_FLOAT_VEC3: TypeInfo("GL_FLOAT_VEC3", GL.GL_FLOAT_VEC3, SIZE_OF_FLOAT * 3),
    GL.GL_FLOAT_VEC4: TypeInfo("GL_FLOAT_VEC4", GL.GL_FLOAT_VEC4, SIZE_OF_FLOAT * 3),
    # Doubles
    GL.GL_DOUBLE: TypeInfo("GL_DOUBLE", GL.GL_DOUBLE, SIZE_OF_DOUBLE),
    GL.GL_DOUBLE_VEC2: TypeInfo("GL_DOUBLE_VEC2", GL.GL_DOUBLE_VEC2, SIZE_OF_DOUBLE * 2),
    GL.GL_DOUBLE_VEC3: TypeInfo("GL_DOUBLE_VEC3", GL.GL_DOUBLE_VEC3, SIZE_OF_DOUBLE * 3),
    GL.GL_DOUBLE_VEC4: TypeInfo("GL_DOUBLE_VEC4", GL.GL_DOUBLE_VEC4, SIZE_OF_DOUBLE * 4),
    # Samplers
    GL.GL_SAMPLER_1D: TypeInfo("GL_SAMPLER_1D", GL.GL_SAMPLER_1D, SIZE_OF_INT),
    GL.GL_SAMPLER_2D: TypeInfo("GL_SAMPLER_2D", GL.GL_SAMPLER_2D, SIZE_OF_INT),
    GL.GL_SAMPLER_3D: TypeInfo("GL_SAMPLER_3D", GL.GL_SAMPLER_3D, SIZE_OF_INT),
    GL.GL_SAMPLER_CUBE: TypeInfo("GL_SAMPLER_CUBE", GL.GL_SAMPLER_CUBE, SIZE_OF_INT),
    GL.GL_SAMPLER_1D_SHADOW: TypeInfo("GL_SAMPLER_1D_SHADOW", GL.GL_SAMPLER_CUBE, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_SHADOW: TypeInfo("GL_SAMPLER_2D_SHADOW", GL.GL_SAMPLER_2D_SHADOW, SIZE_OF_INT),
    GL.GL_SAMPLER_1D_ARRAY: TypeInfo("GL_SAMPLER_1D_ARRAY", GL.GL_SAMPLER_1D_ARRAY, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_ARRAY: TypeInfo("GL_SAMPLER_2D_ARRAY", GL.GL_SAMPLER_2D_ARRAY, SIZE_OF_INT),
    GL.GL_SAMPLER_1D_ARRAY_SHADOW: TypeInfo("GL_SAMPLER_1D_ARRAY_SHADOW", GL.GL_SAMPLER_1D_ARRAY_SHADOW, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_ARRAY_SHADOW: TypeInfo("GL_SAMPLER_2D_ARRAY_SHADOW", GL.GL_SAMPLER_2D_ARRAY_SHADOW, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_MULTISAMPLE: TypeInfo("GL_SAMPLER_2D_MULTISAMPLE", GL.GL_SAMPLER_2D_MULTISAMPLE, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_MULTISAMPLE_ARRAY: TypeInfo("GL_SAMPLER_2D_MULTISAMPLE_ARRAY", GL.GL_SAMPLER_2D_MULTISAMPLE_ARRAY,
                                                 SIZE_OF_INT),
    GL.GL_SAMPLER_CUBE_SHADOW: TypeInfo("GL_SAMPLER_CUBE_SHADOW", GL.GL_SAMPLER_CUBE_SHADOW, SIZE_OF_INT),
    GL.GL_SAMPLER_BUFFER: TypeInfo("GL_SAMPLER_BUFFER", GL.GL_SAMPLER_BUFFER, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_RECT: TypeInfo("GL_SAMPLER_2D_RECT", GL.GL_SAMPLER_2D_RECT, SIZE_OF_INT),
    GL.GL_SAMPLER_2D_RECT_SHADOW: TypeInfo("GL_SAMPLER_2D_RECT_SHADOW", GL.GL_SAMPLER_2D_RECT_SHADOW, SIZE_OF_INT),
    # Integer samplers
    GL.GL_INT_SAMPLER_1D: TypeInfo("GL_INT_SAMPLER_1D", GL.GL_INT_SAMPLER_1D, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_2D: TypeInfo("GL_INT_SAMPLER_2D", GL.GL_INT_SAMPLER_2D, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_3D: TypeInfo("GL_INT_SAMPLER_3D", GL.GL_INT_SAMPLER_3D, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_CUBE: TypeInfo("GL_INT_SAMPLER_CUBE", GL.GL_INT_SAMPLER_CUBE, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_1D_ARRAY: TypeInfo("GL_INT_SAMPLER_1D_ARRAY", GL.GL_INT_SAMPLER_1D_ARRAY, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_2D_ARRAY: TypeInfo("GL_INT_SAMPLER_2D_ARRAY", GL.GL_INT_SAMPLER_2D_ARRAY, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_2D_MULTISAMPLE: TypeInfo("GL_INT_SAMPLER_2D_MULTISAMPLE", GL.GL_INT_SAMPLER_2D_MULTISAMPLE,
                                               SIZE_OF_INT),
    GL.GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY: TypeInfo("GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY",
                                                     GL.GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_BUFFER: TypeInfo("GL_INT_SAMPLER_BUFFER", GL.GL_INT_SAMPLER_BUFFER, SIZE_OF_INT),
    GL.GL_INT_SAMPLER_2D_RECT: TypeInfo("GL_INT_SAMPLER_2D_RECT", GL.GL_INT_SAMPLER_2D_RECT, SIZE_OF_INT),
    # Unsigned integer samplers
    GL.GL_UNSIGNED_INT_SAMPLER_1D: TypeInfo("GL_UNSIGNED_INT_SAMPLER_1D", GL.GL_UNSIGNED_INT_SAMPLER_1D, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_2D: TypeInfo("GL_UNSIGNED_INT_SAMPLER_2D", GL.GL_UNSIGNED_INT_SAMPLER_2D, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_3D: TypeInfo("GL_UNSIGNED_INT_SAMPLER_3D", GL.GL_UNSIGNED_INT_SAMPLER_3D, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_CUBE: TypeInfo("GL_UNSIGNED_INT_SAMPLER_CUBE", GL.GL_UNSIGNED_INT_SAMPLER_CUBE,
                                              SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_1D_ARRAY: TypeInfo("GL_UNSIGNED_INT_SAMPLER_1D_ARRAY",
                                                  GL.GL_UNSIGNED_INT_SAMPLER_1D_ARRAY, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_2D_ARRAY: TypeInfo("GL_UNSIGNED_INT_SAMPLER_2D_ARRAY",
                                                  GL.GL_UNSIGNED_INT_SAMPLER_2D_ARRAY, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE: TypeInfo("GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE",
                                                        GL.GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY: TypeInfo("GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY",
                                                              GL.GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY,
                                                              SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_BUFFER: TypeInfo("GL_UNSIGNED_INT_SAMPLER_BUFFER",
                                                GL.GL_UNSIGNED_INT_SAMPLER_BUFFER, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_SAMPLER_2D_RECT: TypeInfo("GL_UNSIGNED_INT_SAMPLER_2D_RECT",
                                                 GL.GL_UNSIGNED_INT_SAMPLER_2D_RECT, SIZE_OF_INT),
    # Booleans
    GL.GL_BOOL: TypeInfo("GL_BOOL", GL.GL_BOOL, SIZE_OF_INT),
    GL.GL_BOOL_VEC2: TypeInfo("GL_BOOL_VEC2", GL.GL_BOOL_VEC2, SIZE_OF_INT * 2),
    GL.GL_BOOL_VEC3: TypeInfo("GL_BOOL_VEC3", GL.GL_BOOL_VEC3, SIZE_OF_INT * 3),
    GL.GL_BOOL_VEC4: TypeInfo("GL_BOOL_VEC4", GL.GL_BOOL_VEC4, SIZE_OF_INT * 4),
    # Integers
    GL.GL_INT: TypeInfo("GL_INT", GL.GL_INT, SIZE_OF_INT),
    GL.GL_INT_VEC2: TypeInfo("GL_INT_VEC2", GL.GL_INT_VEC2, SIZE_OF_INT * 2),
    GL.GL_INT_VEC3: TypeInfo("GL_INT_VEC3", GL.GL_INT_VEC3, SIZE_OF_INT * 3),
    GL.GL_INT_VEC4: TypeInfo("GL_INT_VEC4", GL.GL_INT_VEC4, SIZE_OF_INT * 4),
    # Unsigned Integers
    GL.GL_UNSIGNED_INT: TypeInfo("GL_UNSIGNED_INT", GL.GL_UNSIGNED_INT, SIZE_OF_INT),
    GL.GL_UNSIGNED_INT_VEC2: TypeInfo("GL_UNSIGNED_INT_VEC2", GL.GL_UNSIGNED_INT_VEC2, SIZE_OF_INT * 2),
    GL.GL_UNSIGNED_INT_VEC3: TypeInfo("GL_UNSIGNED_INT_VEC3", GL.GL_UNSIGNED_INT_VEC3, SIZE_OF_INT * 3),
    GL.GL_UNSIGNED_INT_VEC4: TypeInfo("GL_UNSIGNED_INT_VEC4", GL.GL_UNSIGNED_INT_VEC4, SIZE_OF_INT * 4),
    # Unsigned Short (mostly used for short index buffers)
    GL.GL_UNSIGNED_SHORT: TypeInfo("GL.GL_UNSIGNED_SHORT", GL.GL_UNSIGNED_SHORT, SIZE_OF_SHORT),
    # Byte
    GL.GL_BYTE: TypeInfo("GL_BYTE", GL.GL_BYTE, 1),
    GL.GL_UNSIGNED_BYTE: TypeInfo("GL_UNSIGNED_BYTE", GL.GL_UNSIGNED_BYTE, 1),
    # Matrices (FLOAT)
    GL.GL_FLOAT_MAT2: TypeInfo("GL_FLOAT_MAT2", GL.GL_FLOAT_MAT2, SIZE_OF_FLOAT * 2 * 2),
    GL.GL_FLOAT_MAT3: TypeInfo("GL_FLOAT_MAT3", GL.GL_FLOAT_MAT3, SIZE_OF_FLOAT * 3 * 3),
    GL.GL_FLOAT_MAT4: TypeInfo("GL_FLOAT_MAT4", GL.GL_FLOAT_MAT4, SIZE_OF_FLOAT * 4 * 4),
    GL.GL_FLOAT_MAT2x3: TypeInfo("GL_FLOAT_MAT2x3", GL.GL_FLOAT_MAT2x3, SIZE_OF_FLOAT * 2 * 3),
    GL.GL_FLOAT_MAT2x4: TypeInfo("GL_FLOAT_MAT2x4", GL.GL_FLOAT_MAT2x4, SIZE_OF_FLOAT * 2 * 4),
    GL.GL_FLOAT_MAT3x2: TypeInfo("GL_FLOAT_MAT3x2", GL.GL_FLOAT_MAT3x2, SIZE_OF_FLOAT * 3 * 2),
    GL.GL_FLOAT_MAT3x4: TypeInfo("GL_FLOAT_MAT3x4", GL.GL_FLOAT_MAT3x4, SIZE_OF_FLOAT * 3 * 4),
    GL.GL_FLOAT_MAT4x2: TypeInfo("GL_FLOAT_MAT4x2", GL.GL_FLOAT_MAT4x2, SIZE_OF_FLOAT * 4 * 2),
    GL.GL_FLOAT_MAT4x3: TypeInfo("GL_FLOAT_MAT4x3", GL.GL_FLOAT_MAT4x3, SIZE_OF_FLOAT * 4 * 3),
    # Matrices (DOUBLE)
    GL.GL_DOUBLE_MAT2: TypeInfo("GL_DOUBLE_MAT2", GL.GL_DOUBLE_MAT2, SIZE_OF_DOUBLE * 2 * 2),
    GL.GL_DOUBLE_MAT3: TypeInfo("GL_DOUBLE_MAT3", GL.GL_DOUBLE_MAT3, SIZE_OF_DOUBLE * 3 * 3),
    GL.GL_DOUBLE_MAT4: TypeInfo("GL_DOUBLE_MAT4", GL.GL_DOUBLE_MAT4, SIZE_OF_DOUBLE * 4 * 4),
    GL.GL_DOUBLE_MAT2x3: TypeInfo("GL_DOUBLE_MAT2x3", GL.GL_DOUBLE_MAT2x3, SIZE_OF_DOUBLE * 2 * 3),
    GL.GL_DOUBLE_MAT2x4: TypeInfo("GL_DOUBLE_MAT2x4", GL.GL_DOUBLE_MAT2x4, SIZE_OF_DOUBLE * 2 * 4),
    GL.GL_DOUBLE_MAT3x2: TypeInfo("GL_DOUBLE_MAT3x2", GL.GL_DOUBLE_MAT3x2, SIZE_OF_DOUBLE * 3 * 2),
    GL.GL_DOUBLE_MAT3x4: TypeInfo("GL_DOUBLE_MAT3x4", GL.GL_DOUBLE_MAT3x4, SIZE_OF_DOUBLE * 3 * 4),
    GL.GL_DOUBLE_MAT4x3: TypeInfo("GL_DOUBLE_MAT4x3", GL.GL_DOUBLE_MAT4x3, SIZE_OF_DOUBLE * 4 * 3),
}
